/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.tacdemo

import groovy.util.slurpersupport.GPathResult

class AgentStrategyService {

  boolean transactional = true
  def submissionService

  /*
   * Each onReceived method passes the already converted domain object of the received notification.
   * These objects contain all the information of the server actions you need to implement your strategy.
   * You can also use the raw GPathResult object: a parsed xml dom object of the received notification (parsed with the XmlSlurper)
   */

  def onCashUpdateReceived(CashPosition cashPositionInstance, GPathResult gpathSource) {
    log.info "New Cash update received: .${cashPositionInstance}"
  }

  def onDepotUpdateReceived(DepotPosition depotPositionInstance, GPathResult gpathSource) {
    log.info "New Depot update received: ${depotPositionInstance}"
  }

  def onForecastUpdateReceived(Forecast forecastInstance, GPathResult gpathSource) {
    log.info "New Forecast received: ${forecastInstance}"
    submitDeltaShoutFor (forecastInstance.product)
  }

  def onTradeUpdateReceived(TradeLog tradeLogInstance, GPathResult gpathSource) {
    log.info "New Trade received: ${tradeLogInstance}"
  }

  def onQuoteUpdateReceived(QuoteLog quoteLogInstance, GPathResult gpathSource) {
    log.info "New Quote received: ${quoteLogInstance}"
  }

  def onOrderbookUpdateReceived(Orderbook orderbookInstance, GPathResult gpathSource) {
    log.info "New Orderbook received: ${orderbookInstance}"
  }

  def onProductUpdateReceived(Product productInstance, GPathResult gpathSource) {
    log.info "Product received: ${productInstance} with action ${productInstance.action}"
    submitDeltaShoutFor (productInstance)
  }

  def onOrderStatusUpdateReceived(Shout shout, GPathResult gpathSource) {
    log.info "New Order status update received for shout ${shout}"
  }

  def onNewCompetitionTimeReceived(Date dateTime, Competition competitionInstance) {
    log.info "New Competition Time received: ${dateTime}"
  }

  def onCompetitionReceived(Competition competitionInstance, GPathResult gpathSource) {
    log.info "Competition received: ${competitionInstance} with action ${competitionInstance.action}"
  }


  def submitDeltaShoutFor(Product productInstance) {
    if (!productInstance || !productInstance.enabled) {
      log.info ("Cannot create shout for product ${productInstance} with enabled state ${productInstance?.enabled}")
      return
    }
    def currentForecast = Forecast.findByProduct(productInstance, [sort: 'competitionTime', order: 'desc'])?.forecastValue
    if (!currentForecast) {
      log.info "No current forecast found for product ${productInstance}"
      return
    }
    def currentStock = DepotPosition.findByProductAndLatest(productInstance, true)?.balance ?: 0d
    def delta = Math.abs(currentStock - currentForecast)
    if (delta) {
      Shout shout = new Shout(product: productInstance,
          remainingQuantity: delta,
          limitPrice: 10d,
          buySellIndicator: currentStock < currentForecast ? BuySellIndicator.BUY : BuySellIndicator.SELL,
          orderType: OrderType.LIMIT
      )
      submissionService.submitShout(shout)
      shout.discard()
    }
  }

  protected Double getPrice(Product timeslot) {
    //determine EEX price
    Calendar cal = Calendar.instance
    cal.setTime(timeslot.getStartDateTime())
    double returnValue = 0.0d
    def month = cal.get(Calendar.MONTH)
    def day = cal.get(Calendar.DAY_OF_MONTH)
    def hour = cal.get(Calendar.HOUR_OF_DAY)

    month++  //Adjust month as Java Calendar API starts with month = 0 but profilstore API calculates months starting with 1

    def connUrl = "${}http://ibwmarkets.iw.uni-karlsruhe.de/ps/rest/FYtlyOAmeCmfdVbgkm37oQJsyZ0U1d2u/profile/82?month=${month}&day=${day}&hour=${hour}"
    if (log.isDebugEnabled()) log.debug "${this} tries to retrieve price from: ${connUrl}"
    def xml = new XmlSlurper().parse(connUrl.toURL().getContent())
    if (log.isDebugEnabled()) log.debug "${this} received price data from profilestore: ${xml}"
    if (xml == null) return null
    def value = xml.datapoints.datapoint.find {it.@month == month && it.@day == day && it.@hour == hour && it.@name == this.timeseriesName}?.@value
    if (log.isDebugEnabled()) log.debug "${this} extracted price value from profilestore: ${value}"
    if (value != null && value != '') returnValue = MathUtils.round(value.toDouble() * priceScaling)
    if (log.isDebugEnabled()) log.debug "${this} adjusted price value (scaling factore ${priceScaling}): ${returnValue}"
    return returnValue
  }
}
