/*
 * Copyright 2009-2010 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an
 *
 * "AS IS" BASIS,  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 *
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */

package edu.kit.iism.tacdemo

import groovy.util.slurpersupport.GPathResult

class JmsNotificationService {

  boolean transactional = true
  def agentStrategyService
  def xmlDataBinderService

  def receiveNotification(String xml) throws NullPointerException {
    if (!xml) throw new NullPointerException("receiveNotification triggered with null object")
    def gpathSource = new XmlSlurper().parseText(xml)
    String notification = gpathSource.name()

    switch (notification) {
      case "shoutNotification": orderStatusListener(gpathSource); break
      case "cashPositionNotification": cashPositionListener(gpathSource); break
      case "depotPositionNotification": depotPositionListener(gpathSource); break
      case "forecastNotification": forecastListener(gpathSource);  break
      case "tradeLogNotification": tradeListener(gpathSource);  break
      case "quoteLogNotification": quoteListener(gpathSource); break
      case "orderbookNotification": orderbookListener(gpathSource); break
      case "productNotification": productListener(gpathSource); break
      case "competitionNotification": competitionListener(gpathSource); break
    }
  }

  def orderStatusListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("receiveOrderStatus triggered with null object")

    log.debug "Receiving Orderstatus: ${gpathSource}"

    Shout shout = (Shout) xmlDataBinderService.bindXmlData(Shout.class, gpathSource, ['serverId',
        [competitionId: Competition],
        [productId: Product],
        'buySellIndicator',
        'remainingQuantity',
        'limitPrice',
        'executionQuantity',
        'executionPrice',
        'orderType',
        'dateCreated',
        'dateMod',
        'modReasonCode',
        'outdated',
        'transactionID',
        'shoutID',
        'auctionTradeFlag'
    ])

    def existingShout = Shout.findByServerId(shout.serverId)
    if (existingShout != null) {
      existingShout.copyPropertiesFrom(shout)
      if (!existingShout.save(flush: true)) log.error "Failed to save outdated shout update: ${shout?.errors}"
    } else {
      if (!shout.save(flush: true)) {
        log.error "Failed to save received shout update: ${shout?.errors}"
      } else {
        agentStrategyService.onOrderStatusUpdateReceived(shout, gpathSource)
      }
    }
    return null
  }

  def cashPositionListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("receiveCashMap triggered with null object")
    log.debug "Receiving Cash update: ${gpathSource}"
    CashPosition cashPosition = (CashPosition) xmlDataBinderService.bindXmlData(CashPosition.class, gpathSource, ['serverId',
        [competitionId: Competition],
        'change',
        'balance',
        'description',
        'dateCreated',
        'transactionID',
        'latest'])

    def oldCashPosition = CashPosition.findByLatest(true)
    if (oldCashPosition) {
      oldCashPosition.latest = false
      if (!oldCashPosition.save()) {
        log.error "Failed to save old cashPosition as outdated: ${oldCashPosition?.errors}"
      }
    }

    if (!cashPosition.save(flush: true)) {
      log.error "Failed to save received cashPositionNotification: ${cashPosition?.errors}"
    } else {
      agentStrategyService.onCashUpdateReceived(cashPosition, gpathSource)
    }
    return null
  }

  def depotPositionListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("receiveDepot triggered with null object")
    log.debug "Receiving Depot Update: ${gpathSource}"
    DepotPosition depotPosition = (DepotPosition) xmlDataBinderService.bindXmlData(DepotPosition.class, gpathSource, ['serverId',
        [competitionId: Competition],
        [productId: Product],
        'change',
        'balance',
        'description',
        'dateCreated',
        'transactionID',
        'latest'])

    def oldDepotPosition = DepotPosition.findByLatest(true)
    if (oldDepotPosition) {
      oldDepotPosition.latest = false
      if (!oldDepotPosition.save()) {
        log.error "Failed to save old cashPosition as outdated: ${oldDepotPosition?.errors}"
      }
    }

    if (!depotPosition.save(flush: true)) {
      log.error "Failed to save received depotPositionNotification: ${depotPosition?.errors}"
    } else {
      agentStrategyService.onDepotUpdateReceived(depotPosition, gpathSource)
    }
    return null
  }

  def forecastListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("receiveForecast triggered with null object")
    log.debug "Receiving Forecast: ${gpathSource}"

    Forecast forecast = (Forecast) xmlDataBinderService.bindXmlData(Forecast.class, gpathSource, ['serverId',
        [competitionId: Competition],
        [productId: Product],
        'competitionTime',
        'forecastValue'])
    if (!forecast.save(flush: true)) {
      log.error "Failed to save received forecastNotification as local Forecast: ${forecast?.errors}"
    } else {
      agentStrategyService.onForecastUpdateReceived(forecast, gpathSource)
    }
    return null
  }

  /**
   * Listener subscribes to topic <code>public.Trade</code>
   * receives TradeNotification map and saves it as local TradeLog object in database
   */
  def tradeListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("tradeListener triggered with null object")
    log.debug "Receiving Trade: ${gpathSource}"

    TradeLog tradeLog = (TradeLog) xmlDataBinderService.bindXmlData(TradeLog.class, gpathSource, ['serverId',
        [competitionId: Competition],
        [productId: Product],
        'price',
        'quantity'])

    if (!tradeLog.save(flush: true)) {
      log.error "Failed to save received TradeNotification: ${tradeLog?.errors}"
    } else {
      agentStrategyService.onTradeUpdateReceived(tradeLog, gpathSource)
    }
    return null
  }

  /*
  * Listener subscribes to topic "public.Trade"
  * receives TradeNotification map and saves it as local TradeLog object in database
  */

  def quoteListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("quoteListener triggered with null object")
    log.debug "Receiving Quote: ${gpathSource}"
    QuoteLog quoteLog = (QuoteLog) xmlDataBinderService.bindXmlData(QuoteLog.class, gpathSource, ['serverId',
        [competitionId: Competition],
        [productId: Product],
        'bid',
        'bidSize',
        'ask',
        'askSize'])
    if (!quoteLog.save(flush: true)) {
      log.error "Failed to save received QuoteNotification: ${quoteLog?.errors}"
    } else {
      agentStrategyService.onQuoteUpdateReceived(quoteLog, gpathSource)
    }
    return null
  }

  def orderbookListener(GPathResult gpathSource) throws NullPointerException {

    if (!gpathSource) throw new NullPointerException("orderbookListener triggered with null object")
    log.debug "Receiving Orderbook: ${gpathSource}"
    Orderbook orderbook = (Orderbook) xmlDataBinderService.bindXmlData(Orderbook.class, gpathSource, ['serverId',
        [competitionId: Competition],
        [productId: Product],
        'dateExecuted',
        'transactionID',
        'outdated',
        'bid1',
        'bid2',
        'bid3',
        'bid4',
        'bid5',
        'bid6',
        'bid7',
        'bid8',
        'bid9',
        'bid10',
        'bidSize1',
        'bidSize2',
        'bidSize3',
        'bidSize4',
        'bidSize5',
        'bidSize6',
        'bidSize7',
        'bidSize8',
        'bidSize9',
        'bidSize10',
        'ask1',
        'ask2',
        'ask3',
        'ask4',
        'ask5',
        'ask6',
        'ask7',
        'ask8',
        'ask9',
        'ask10',
        'askSize1',
        'askSize2',
        'askSize3',
        'askSize4',
        'askSize5',
        'askSize6',
        'askSize7',
        'askSize8',
        'askSize9',
        'askSize10'])

    def oldOrderbook = Orderbook.findByOutdatedAndProduct(false, orderbook?.product)
    if (oldOrderbook) {
      oldOrderbook.outdated = true
      if (!oldOrderbook.save()) {
        log.error "Failed to save old cashPosition as outdated: ${oldOrderbook?.errors}"
      }
    }

    if (!orderbook.save(flush: true)) {
      log.error "Failed to save received OrderbookNotification: ${orderbook?.errors}"
    } else {
      agentStrategyService.onOrderbookUpdateReceived(orderbook, gpathSource)
    }
    return null
  }

  /*
  * Listener subscribes to topic "public.Products"
  * receives Product object and status message = ("Product opened" | "Product closed")
  */

  def productListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("productListener triggered with null object")
    log.debug "Received Product Information: ${gpathSource}"
    Product incomingProduct = (Product) xmlDataBinderService.bindXmlData(Product.class, gpathSource, ['serverId',
        [competitionId: Competition],
        'name',
        'enabled',
        'startDateTime',
        'endDateTime',
        'serialNumber',
        'current',
        'action'])


    def isValid = incomingProduct?.validate()
    if (!isValid) {
      log.error "Failed to validate product: ${incomingProduct?.errors}"
      incomingProduct?.discard()
    } else {
      def existingProduct = Product.findByServerId(incomingProduct.serverId)
      if (existingProduct) {
        log.debug "Existing product found: ${existingProduct}... Updating properties..."
        existingProduct.copyPropertiesFrom (incomingProduct)
        incomingProduct.discard()
        if (!existingProduct.save(flush: true)) {
          log.error ("Error saving existing product: ${existingProduct.errors}")
        } else {
          agentStrategyService.onProductUpdateReceived(existingProduct, gpathSource)
        }
      } else {
        log.debug "No existing product found in db for ${incomingProduct}. Saving incoming record as new product."
        if (!incomingProduct.save(flush: true)) {
          log.error ("Error saving new product: ${incomingProduct.errors}")
        } else {
          agentStrategyService.onProductUpdateReceived(incomingProduct, gpathSource)
        }
      }
    }
    return null
  }

/*
* Listener subscribes to topic "public.CompetitionAnnounce"
*/

  def competitionListener(GPathResult gpathSource) throws NullPointerException {
    if (!gpathSource) throw new NullPointerException("competitionListener triggered with null object as input")

    log.debug "Receiving Competition Announcement: ${gpathSource}"
    Competition incomingCompetition = (Competition) xmlDataBinderService.bindXmlData(Competition.class, gpathSource,
        ['serverId',
            'name',
            'competitionStatus',
            'description',
            'initialDuration',
            'durationBetweenShifts',
            'timeslotLength',
            'timeslotsOverall',
            'timeslotsOpen',
            'deactivateTimeslotsAhead',
            'simulationStartTime',
            'participantCount',
            'balancingCostOver',
            'balancingCostUnder',
            'action',
            'current',
            'currentCompetitionTime'
        ])

    def isValid = incomingCompetition?.validate()
    log.debug "Competition Announcement valid: ${isValid}"
    if (!isValid) {
      log.error "Failed to validate competition: ${incomingCompetition?.errors}"
      incomingCompetition?.discard()
    } else {
      log.debug "Trying to find existing competition based on serverId ${incomingCompetition.serverId} from incoming competitionAnnouncement"
      def existingCompetition = Competition.findByServerId(incomingCompetition.serverId)
      if (existingCompetition != null) {
        log.debug "Existing competition found: ${existingCompetition}... Updating properties..."
        existingCompetition.copyPropertiesFrom (incomingCompetition)
        incomingCompetition.discard() //discard new one as its properties were copied and saved to an existing instance
        if (!existingCompetition.save(flush: true)){
          log.error "Error saving updated competition: ${existingCompetition.errors}"
        } else {
          if (existingCompetition.action == ActionType.TimeAdvanced) {
            agentStrategyService.onNewCompetitionTimeReceived(existingCompetition.currentCompetitionTime, existingCompetition)
          } else {
            agentStrategyService.onCompetitionReceived(existingCompetition, gpathSource)
          }

        }
      } else {
        log.debug "No existing competition found in db for ${incomingCompetition}. Saving incoming record as new competition."
        if (!incomingCompetition.save(flush: true)) log.error "Error saving new competition: ${incomingCompetition.errors}"
        existingCompetition = incomingCompetition
        agentStrategyService.onCompetitionReceived(existingCompetition, gpathSource)
      }

    }
    return null
  }
}
